"""Core utilities"""
import typing
import logging
import os
import hashlib
import io
import shutil
import tarfile
import zipfile
import urllib.parse
import urllib.request
import urllib.error

import six
from tqdm import tqdm

log = logging.getLogger(__name__)

# Progbar, _extract_archive, get_file, _hash_file and validate_file
# adapted / copied from Keras to avoid requiring
# TensorFlow to simply use file downloads, etc.

# ProgressFileObject adapted from https://stackoverflow.com/a/3668977


def get_datadir_base():
    """Get the base directory for data."""
    cache_dir_env = os.environ.get("MIRA_CACHE")
    if cache_dir_env is not None:
        cache_dir = os.path.abspath(cache_dir_env)
    else:
        cache_dir = os.path.join(os.path.expanduser("~"), ".mira")
    return os.path.expanduser(cache_dir)


class ProgressFileObject(io.FileIO):  # noqa: E302
    """A file downloader that reports progress."""

    def __init__(self, path, *args, **kwargs):
        self._total_size = os.path.getsize(path)
        self.progbar = tqdm(total=self._total_size, desc="Reading " + path)
        io.FileIO.__init__(self, path, *args, **kwargs)

    def read(self, size=-1):
        if size > 0:
            self.progbar.update(size)
        return io.FileIO.read(self, size)

    def close(self):
        self.progbar.close()
        super().close()


# pylint: disable=consider-using-with
def _extract_archive(  # noqa: E302
    file_path, path=".", archive_format="auto", extract_check_fn=None
):
    """Extracts an archive if it matches tar, tar.gz, tar.bz, or zip formats.

    # Arguments
        file_path: path to the archive file
        path: path to extract the archive file
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    # Returns
        True if a match was found and an archive extraction was completed,
        False otherwise.
    """
    if archive_format is None:
        return False
    if archive_format == "auto":
        archive_format = ["tar", "zip"]
    if isinstance(archive_format, six.string_types):
        archive_format = [archive_format]

    for archive_type in archive_format:
        open_fn: typing.Callable[
            [typing.Any], typing.Union[tarfile.TarFile, zipfile.ZipFile]
        ]
        is_match_fn: typing.Callable[[typing.Any], bool]
        if archive_type == "tar":
            # Implement progress bar for tar files,
            # per https://stackoverflow.com/questions/3667865/python-tarfile-progress-output  # pylint: disable=line-too-long
            # open_fn = tarfile.open
            open_fn = lambda fp: tarfile.open(
                fileobj=ProgressFileObject(fp)
            )  # noqa: E731,E501
            is_match_fn = tarfile.is_tarfile
        if archive_type == "zip":
            # open_fn = zipfile.ZipFile
            open_fn = lambda fp: zipfile.ZipFile(
                file=ProgressFileObject(fp)
            )  # noqa: E731,E501
            is_match_fn = zipfile.is_zipfile

        if is_match_fn(file_path):
            if os.path.exists(path) and extract_check_fn is not None:
                log.info(
                    "Found target folder. Checking for " "existing complete extraction."
                )
                if extract_check_fn(path):
                    log.info("Found complete extraction.")
                    return True
                log.info("Extraction is not complete. Re-extracting.")
            log.info("Extracting %s to %s", file_path, path)
            with open_fn(file_path) as archive:
                try:
                    archive.extractall(path)
                except (tarfile.TarError, RuntimeError, KeyboardInterrupt):
                    if os.path.exists(path):
                        if os.path.isfile(path):
                            os.remove(path)
                        else:
                            shutil.rmtree(path)
                    raise
            return True
    return False


def get_file(
    origin,
    fname=None,
    file_hash=None,
    cache_subdir="datasets",
    hash_algorithm="auto",
    extract=False,
    archive_format="auto",
    extract_check_fn=None,
):
    """Downloads a file from a URL if it not already in the cache.

    By default the file at the url `origin` is downloaded to the
    cache_dir `~/.keras`, placed in the cache_subdir `datasets`,
    and given the filename `fname`. The final location of a file
    `example.txt` would therefore be `~/.keras/datasets/example.txt`.

    Files in tar, tar.gz, tar.bz, and zip formats can also be extracted.
    Passing a hash will verify the file after download. The command line
    programs `shasum` and `sha256sum` can compute the hash.

    # Arguments
        fname: Name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location. If None,
            the name of the file at origin is used instead.
        origin: Original URL of the file.
        untar: Deprecated in favor of 'extract'.
            boolean, whether the file should be decompressed
        file_hash: The expected hash string of the file after download.
            The sha256 and md5 hash algorithms are both supported.
        cache_subdir: Subdirectory under the Keras cache dir where the file is
            saved. If an absolute path `/path/to/folder` is
            specified the file will be saved at that location.
        hash_algorithm: Select the hash algorithm to verify the file.
            options are 'md5', 'sha256', and 'auto'.
            The default 'auto' detects the hash algorithm in use.
        extract: True tries extracting the file as an Archive, like tar or zip.
        archive_format: Archive format to try for extracting the file.
            Options are 'auto', 'tar', 'zip', and None.
            'tar' includes tar, tar.gz, and tar.bz files.
            The default 'auto' is ['tar', 'zip'].
            None or an empty list will return no matches found.

    # Returns
        Path to the downloaded file
    """
    if fname is None:
        fname = os.path.basename(urllib.parse.urlparse(origin).path)
    datadir_base = get_datadir_base()
    datadir = os.path.join(datadir_base, cache_subdir)
    if not os.path.exists(datadir):
        os.makedirs(datadir)

    fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        # File found; verify integrity if a hash was provided.
        if file_hash is not None:
            if not validate_file(fpath, file_hash, algorithm=hash_algorithm):
                log.warning(
                    "A local file was found, but it seems to be "
                    "incomplete or outdated because the %s"
                    " file hash does not match the original value of %s"
                    " so we will re-download the data.",
                    hash_algorithm,
                    file_hash,
                )
                download = True
    else:
        download = True

    if download:
        error_msg = "URL fetch failure on {} : {}"
        try:
            try:
                urllib.request.urlretrieve(origin, fpath)
            except urllib.error.HTTPError as e:  # pylint: disable=invalid-name
                raise Exception(error_msg.format(origin, e.code)) from e
            except urllib.error.URLError as e:  # pylint: disable=invalid-name
                raise Exception(error_msg.format(origin, e.errno)) from e
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise
    if extract:
        extract_target = os.path.join(
            datadir, os.path.split(fpath)[1].split(os.extsep)[0]
        )
        extract_result = _extract_archive(
            file_path=fpath,
            path=extract_target,
            archive_format=archive_format,
            extract_check_fn=extract_check_fn,
        )
        if extract_result:
            fpath = extract_target

    return fpath


def _hash_file(fpath, algorithm="sha256", chunk_size=65535):
    """Calculates a file sha256 or md5 hash.

    # Example

    ```python
        >>> from keras.data_utils import _hash_file
        >>> _hash_file('/path/to/file.zip')
        'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    ```

    # Arguments
        fpath: path to the file being validated
        algorithm: hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        The file hash
    """
    if algorithm in ("sha256", "auto"):
        hasher = hashlib.sha256()
    else:
        hasher = hashlib.md5()

    with open(fpath, "rb") as fpath_file:
        for chunk in iter(lambda: fpath_file.read(chunk_size), b""):
            hasher.update(chunk)

    return hasher.hexdigest()


def validate_file(fpath, file_hash, algorithm="auto", chunk_size=65535):
    """Validates a file against a sha256 or md5 hash.

    # Arguments
        fpath: path to the file being validated
        file_hash:  The expected hash string of the file.
            The sha256 and md5 hash algorithms are both supported.
        algorithm: Hash algorithm, one of 'auto', 'sha256', or 'md5'.
            The default 'auto' detects the hash algorithm in use.
        chunk_size: Bytes to read at a time, important for large files.

    # Returns
        Whether the file is valid
    """
    if (algorithm == "sha256") or (algorithm == "auto" and len(file_hash) == 64):
        hasher = "sha256"
    else:
        hasher = "md5"
    log.info("Checking hash for %s.", os.path.split(fpath)[1])
    return str(_hash_file(fpath, hasher, chunk_size)) == str(file_hash)



""" PyTorch Soft-NMS

This code was adapted from a PR for detectron2 submitted by https://github.com/alekseynp
https://github.com/facebookresearch/detectron2/pull/1183/files

Detectron2 is licensed Apache 2.0, Copyright Facebook Inc.
"""
import torch
from typing import List


def pairwise_iou(boxes1, boxes2) -> torch.Tensor:
    """
    Given two lists of boxes of size N and M,
    compute the IoU (intersection over union)
    between __all__ N x M pairs of boxes.
    The box order must be (xmin, ymin, xmax, ymax).
    Args:
        boxes1,boxes2 (Boxes): two `Boxes`. Contains N & M boxes, respectively.
    Returns:
        Tensor: IoU, sized [N,M].
    """
    area1 = (boxes1[:, 2] - boxes1[:, 0]) * (boxes1[:, 3] - boxes1[:, 1])  # [N,]
    area2 = (boxes2[:, 2] - boxes2[:, 0]) * (boxes2[:, 3] - boxes2[:, 1])  # [M,]

    width_height = torch.min(boxes1[:, None, 2:], boxes2[:, 2:]) - torch.max(
        boxes1[:, None, :2], boxes2[:, :2]
    )  # [N,M,2]

    width_height.clamp_(min=0)  # [N,M,2]
    inter = width_height.prod(dim=2)  # [N,M]

    # handle empty boxes
    iou = torch.where(
        inter > 0,
        inter / (area1[:, None] + area2 - inter),
        torch.zeros(1, dtype=inter.dtype, device=inter.device),
    )
    return iou


def soft_nms(
    boxes,
    scores,
    method_gaussian: bool = True,
    sigma: float = 0.5,
    iou_threshold: float = .5,
    score_threshold: float = 0.005
):
    """
    Soft non-max suppression algorithm.

    Implementation of [Soft-NMS -- Improving Object Detection With One Line of Codec]
    (https://arxiv.org/abs/1704.04503)

    Args:
        boxes_remain (Tensor[N, ?]):
           boxes where NMS will be performed
           if Boxes, in (x1, y1, x2, y2) format
           if RotatedBoxes, in (x_ctr, y_ctr, width, height, angle_degrees) format
        scores_remain (Tensor[N]):
           scores for each one of the boxes
        method_gaussian (bool): use gaussian method if True, otherwise linear        
        sigma (float):
           parameter for Gaussian penalty function
        iou_threshold (float):
           iou threshold for applying linear decay. Nt from the paper
           re-used as threshold for standard "hard" nms
        score_threshold (float):
           boxes with scores below this threshold are pruned at each iteration.
           Dramatically reduces computation time. Authors use values in [10e-4, 10e-2]

    Returns:
        tuple(Tensor, Tensor):
            [0]: int64 tensor with the indices of the elements that have been kept
            by Soft NMS, sorted in decreasing order of scores
            [1]: float tensor with the re-scored scores of the elements that were kept
    """
    device = boxes.device
    boxes_remain = boxes.clone()
    scores_remain = scores.clone()
    num_elem = scores_remain.size()[0]
    idxs = torch.arange(num_elem)
    idxs_out = torch.zeros(num_elem, dtype=torch.int64, device=device)
    scores_out = torch.zeros(num_elem, dtype=torch.float32, device=device)
    count: int = 0

    while scores_remain.numel() > 0:
        top_idx = torch.argmax(scores_remain)
        idxs_out[count] = idxs[top_idx]
        scores_out[count] = scores_remain[top_idx]
        count += 1

        top_box = boxes_remain[top_idx]
        ious = pairwise_iou(top_box.unsqueeze(0), boxes_remain)[0]

        if method_gaussian:
            decay = torch.exp(-torch.pow(ious, 2) / sigma)
        else:
            decay = torch.ones_like(ious)
            decay_mask = ious > iou_threshold
            decay[decay_mask] = 1 - ious[decay_mask]

        scores_remain *= decay
        keep = scores_remain > score_threshold
        keep[top_idx] = torch.tensor(False, device=device)

        boxes_remain = boxes_remain[keep]
        scores_remain = scores_remain[keep]
        idxs = idxs[keep]

    return idxs_out[:count], scores_out[:count]


def batched_soft_nms(
    boxes, scores, idxs,
    method_gaussian: bool = True,
    sigma: float = 0.5,
    iou_threshold: float = .5,
    score_threshold: float = 0.001):

    """
    Performs soft non-maximum suppression in a batched fashion.

    Each index value correspond to a category, and NMS
    will not be applied between elements of different categories.

    Args:
        boxes (Tensor[N, 4]):
           boxes where NMS will be performed. They
           are expected to be in (x1, y1, x2, y2) format
        scores (Tensor[N]):
           scores for each one of the boxes
        idxs (Tensor[N]):
           indices of the categories for each one of the boxes.
        method (str):
           one of ['gaussian', 'linear', 'hard']
           see paper for details. users encouraged not to use "hard", as this is the
           same nms available elsewhere in detectron2
        sigma (float):
           parameter for Gaussian penalty function
        iou_threshold (float):
           iou threshold for applying linear decay. Nt from the paper
           re-used as threshold for standard "hard" nms
        score_threshold (float):
           boxes with scores below this threshold are pruned at each iteration.
           Dramatically reduces computation time. Authors use values in [10e-4, 10e-2]
    Returns:
        tuple(Tensor, Tensor):
            [0]: int64 tensor with the indices of the elements that have been kept
            by Soft NMS, sorted in decreasing order of scores
            [1]: float tensor with the re-scored scores of the elements that were kept
    """
    if boxes.numel() == 0:
        return (
            torch.empty((0,), dtype=torch.int64, device=boxes.device),
            torch.empty((0,), dtype=torch.float32, device=scores.device),
        )
    # strategy: in order to perform NMS independently per class.
    # we add an offset to all the boxes. The offset is dependent
    # only on the class idx, and is large enough so that boxes
    # from different classes do not overlap
    max_coordinate = boxes.max()
    offsets = idxs.to(boxes) * (max_coordinate + 1)
    boxes_for_nms = boxes + offsets[:, None]
    return soft_nms(
        boxes_for_nms, scores, method_gaussian=method_gaussian, sigma=sigma,
        iou_threshold=iou_threshold, score_threshold=score_threshold
    )




from .preloaded import (  # noqa: F401
    load_voc2012,
    load_oxfordiiitpets,
    load_shapes,
    load_random_images,
    COCOAnnotationConfiguration,
    VOCAnnotationConfiguration,
    COCOAnnotationConfiguration90,
)
from .coco import load_coco  # noqa: F401
from .voc import load_voc  # noqa: F401



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mira',
 'mira.core',
 'mira.core.protos',
 'mira.datasets',
 'mira.detectors',
 'mira.detectors.assets.serve',
 'mira.detectors.experimental',
 'mira.thirdparty',
 'mira.thirdparty.detr',
 'mira.thirdparty.detr.models',
 'mira.thirdparty.detr.util',
 'mira.thirdparty.effdet',
 'mira.thirdparty.effdet.config',
 'mira.thirdparty.effdet.object_detection']

package_data = \
{'': ['*'], 'mira.datasets': ['assets/*']}

install_requires = \
['albumentations',
 'matplotlib',
 'numpy',
 'pandas',
 'protobuf',
 'tqdm',
 'validators']

extras_require = \
{'detectors': ['timm', 'omegaconf', 'torchvision']}

setup_kwargs = {
    'name': 'mira',
    'version': '0.5.30',
    'description': 'A package for simplifying object detection',
    'long_description': '# mira [![CircleCI](https://circleci.com/gh/faustomorales/mira.svg?style=shield)](https://circleci.com/gh/faustomorales/mira) [![Documentation Status](https://readthedocs.org/projects/mira-python/badge/?version=latest)](https://mira-python.readthedocs.io/en/latest/?badge=latest)\n\nmira provides tooling for simple object detection projects. The package spans three areas of focus.\n\n- **Core** object detection abstractions for images and annotations\n- Access to **datasets** from common formats (e.g., VOC, COCO) and image sets (e.g., VOC 2012)\n- A common API to for **well-known models** (e.g., EfficientDet and FasterRCNN)\n\nCheck out [the docs](https://mira-python.readthedocs.io/en/latest/).\n\n## Installation\n\n```shell\npip install mira\n```\n',
    'author': 'Fausto Morales',
    'author_email': 'faustomorales@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/faustomorales/mira',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.2,<3.11',
}


setup(**setup_kwargs)



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Bounding Box List definition.

BoxList represents a list of bounding boxes as tensorflow
tensors, where each bounding box is represented as a row of 4 numbers,
[y_min, x_min, y_max, x_max].  It is assumed that all bounding boxes
within a given list correspond to a single image.  See also
box_list.py for common box related operations (such as area, iou, etc).

Optionally, users can add additional related fields (such as weights).
We assume the following things to be true about fields:
* they correspond to boxes in the box_list along the 0th dimension
* they have inferable rank at graph construction time
* all dimensions except for possibly the 0th can be inferred
  (i.e., not None) at graph construction time.

Some other notes:
    * Following tensorflow conventions, we use height, width ordering,
        and correspondingly, y,x (or ymin, xmin, ymax, xmax) ordering
    * Tensors are always provided as (flat) [N, 4] tensors.
"""
import torch
from typing import Optional, List, Dict


@torch.jit.script
class BoxList(object):
    """Box collection."""
    data: Dict[str, torch.Tensor]

    def __init__(self, boxes):
        """Constructs box collection.

        Args:
            boxes: a tensor of shape [N, 4] representing box corners

        Raises:
            ValueError: if invalid dimensions for bbox data or if bbox data is not in float32 format.
        """
        if len(boxes.shape) != 2 or boxes.shape[-1] != 4:
            raise ValueError('Invalid dimensions for box data.')
        if boxes.dtype != torch.float32:
            raise ValueError('Invalid tensor type: should be tf.float32')
        self.data = {'boxes': boxes}

    def num_boxes(self):
        """Returns number of boxes held in collection.

        Returns:
          a tensor representing the number of boxes held in the collection.
        """
        return self.data['boxes'].shape[0]

    def get_all_fields(self):
        """Returns all fields."""
        return self.data.keys()

    def get_extra_fields(self):
        """Returns all non-box fields (i.e., everything not named 'boxes')."""
        # return [k for k in self.data.keys() if k != 'boxes']  # FIXME torscript doesn't support comprehensions yet
        extra: List[str] = []
        for k in self.data.keys():
            if k != 'boxes':
                extra.append(k)
        return extra

    def add_field(self, field: str, field_data: torch.Tensor):
        """Add field to box list.

        This method can be used to add related box data such as weights/labels, etc.

        Args:
            field: a string key to access the data via `get`
            field_data: a tensor containing the data to store in the BoxList
        """
        self.data[field] = field_data

    def has_field(self, field: str):
        return field in self.data

    #@property  # FIXME for torchscript compat
    def boxes(self):
        """Convenience function for accessing box coordinates.

        Returns:
            a tensor with shape [N, 4] representing box coordinates.
        """
        return self.get_field('boxes')

    #@boxes.setter  # FIXME for torchscript compat
    def set_boxes(self, boxes):
        """Convenience function for setting box coordinates.

        Args:
            boxes: a tensor of shape [N, 4] representing box corners

        Raises:
            ValueError: if invalid dimensions for bbox data
        """
        if len(boxes.shape) != 2 or boxes.shape[-1] != 4:
            raise ValueError('Invalid dimensions for box data.')
        self.data['boxes'] = boxes

    def get_field(self, field: str):
        """Accesses a box collection and associated fields.

        This function returns specified field with object; if no field is specified,
        it returns the box coordinates.

        Args:
            field: this optional string parameter can be used to specify a related field to be accessed.

        Returns:
            a tensor representing the box collection or an associated field.

        Raises:
            ValueError: if invalid field
        """
        if not self.has_field(field):
            raise ValueError(f'field {field} does not exist')
        return self.data[field]

    def set_field(self, field: str, value: torch.Tensor):
        """Sets the value of a field.

        Updates the field of a box_list with a given value.

        Args:
            field: (string) name of the field to set value.
            value: the value to assign to the field.

        Raises:
            ValueError: if the box_list does not have specified field.
        """
        if not self.has_field(field):
            raise ValueError(f'field {field} does not exist')
        self.data[field] = value

    def get_center_coordinates_and_sizes(self):
        """Computes the center coordinates, height and width of the boxes.

        Returns:
            a list of 4 1-D tensors [ycenter, xcenter, height, width].
        """
        box_corners = self.boxes()
        ymin, xmin, ymax, xmax = box_corners.t().unbind()
        width = xmax - xmin
        height = ymax - ymin
        ycenter = ymin + height / 2.
        xcenter = xmin + width / 2.
        return [ycenter, xcenter, height, width]

    def transpose_coordinates(self):
        """Transpose the coordinate representation in a boxlist.

        """
        y_min, x_min, y_max, x_max = self.boxes().chunk(4, dim=1)
        self.set_boxes(torch.cat([x_min, y_min, x_max, y_max], 1))

    def as_tensor_dict(self, fields: Optional[List[str]] = None):
        """Retrieves specified fields as a dictionary of tensors.

        Args:
            fields: (optional) list of fields to return in the dictionary.
                If None (default), all fields are returned.

        Returns:
            tensor_dict: A dictionary of tensors specified by fields.

        Raises:
            ValueError: if specified field is not contained in boxlist.
        """
        tensor_dict = {}
        if fields is None:
            fields = self.get_all_fields()
        for field in fields:
            if not self.has_field(field):
                raise ValueError('boxlist must contain all specified fields')
            tensor_dict[field] = self.get_field(field)
        return tensor_dict

    #@property
    def device(self):
        return self.data['boxes'].device



from omegaconf import OmegaConf


def default_detection_train_config():
    # FIXME currently using args for train config, will revisit, perhaps move to Hydra
    h = OmegaConf.create()

    # dataset
    h.skip_crowd_during_training = True

    # augmentation
    h.input_rand_hflip = True
    h.train_scale_min = 0.1
    h.train_scale_max = 2.0
    h.autoaugment_policy = None

    # optimization
    h.momentum = 0.9
    h.learning_rate = 0.08
    h.lr_warmup_init = 0.008
    h.lr_warmup_epoch = 1.0
    h.first_lr_drop_epoch = 200.0
    h.second_lr_drop_epoch = 250.0
    h.clip_gradients_norm = 10.0
    h.num_epochs = 300

    # regularization l2 loss.
    h.weight_decay = 4e-5

    h.lr_decay_method = 'cosine'
    h.moving_average_decay = 0.9998
    h.ckpt_var_scope = None

    return h



import typing
import itertools

import cv2
import numpy as np
import scipy.sparse.csgraph as ssc

from . import utils


def collapse_boxes(boxes: np.ndarray, threshold=0.5, mode="smallest"):
    """Given a set of boxes, collapse overlapping boxes into the smallest
    or largest common area."""
    assert mode in ["smallest", "largest"], f"Unknown mode: {mode}"
    n_components, labels = ssc.connected_components(
        utils.compute_iou(boxes, boxes) > threshold, directed=False
    )
    return np.array(
        [
            (
                [subgroup[:, :2].max(axis=0), subgroup[:, 2:].min(axis=0)]
                if (
                    (subgroup[:, 2:].min(axis=0) - subgroup[:, :2].max(axis=0)) > 0
                ).all()
                else subgroup[
                    np.product(subgroup[:, 2:] - subgroup[:, :2], axis=1).argmin()
                ]
            )
            if mode == "smallest"
            else [subgroup[:, :2].min(axis=0), subgroup[:, 2:].max(axis=0)]
            for subgroup in [
                boxes[labels == component] for component in range(n_components)
            ]
        ]
    ).reshape(-1, 4)


def find_consensus_regions(
    bbox_groups: typing.List[np.ndarray], iou_threshold: float = 0.5
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Given a set of box groups for the same image being
    labeled by different people, find the regions of
    consensus and non-consensus."""
    exclude_list = []
    for (bboxes1, bboxes2), annIdx in itertools.product(
        itertools.combinations(bbox_groups, 2),
        range(max([g[:, -1].max() if len(g) > 0 else 0 for g in bbox_groups]) + 1),
    ):
        bboxes1, bboxes2 = [b[b[:, -1] == annIdx, :-1] for b in [bboxes1, bboxes2]]
        if not len(bboxes1) > 0 and not len(bboxes2) > 0:
            continue
        if len(bboxes1) > 0 and not len(bboxes2) > 0:
            exclude_list.extend(bboxes1)
        elif not len(bboxes1) > 0 and len(bboxes2) > 0:
            exclude_list.extend(bboxes2)
        else:
            iou = utils.compute_iou(bboxes1, bboxes2)
            exclude_list.extend(bboxes1[~(iou.max(axis=1) > iou_threshold)])
            exclude_list.extend(bboxes2[~(iou.max(axis=0) > iou_threshold)])
    exclude = (
        np.array(exclude_list)
        if len(exclude_list) > 0
        else np.empty((0, 4), dtype="int64")
    )
    include = np.concatenate(
        [
            g[:, :-1] if len(g) > 0 else np.empty((0, 4), dtype="int64")
            for g in bbox_groups
        ],
        axis=0,
    )
    if len(include) > 0:
        include = np.unique(
            include[utils.compute_iou(include, exclude).max(axis=1) == 0]
            if len(exclude) > 0
            else include,
            axis=0,
        )
    return include, exclude


def search(x, y, exclude, include, max_height, max_width, min_height=1, min_width=1):
    """Starting at some coordinates, search for the largest rectangle
    that encompasses include boxes fully and does not cross any exclude boxes."""
    crossings = np.concatenate(
        [
            exclude[
                (
                    (x >= exclude[:, 0])
                    & (x < exclude[:, 2])
                    & (y >= exclude[:, 1])
                    & (y < exclude[:, 3])
                )
            ],
            include[
                (
                    (x > include[:, 0])
                    & (x < include[:, 2])
                    & (y > include[:, 1])
                    & (y < include[:, 3])
                )
            ],
        ],
        axis=0,
    )
    if len(crossings) > 0:
        return (1, crossings[:, 3].max() - y, False)
    xye = exclude[(exclude[:, 2:] > (x, y)).min(axis=1)]
    xyi = include[(include[:, 2:] > (x, y)).min(axis=1)]

    # We can extend as far as (a) the first inclusion box that
    # crosses we would split vertically or (b) the first exclusion
    # box that we would split vertically or (c) the maximum height or
    # (d) the image height.
    dyc_max = int(
        np.concatenate(
            [xyi[xyi[:, 0] < x, 1] - y, xye[xye[:, 0] <= x, 1] - y, [max_height]]
        ).min()
    )
    dx, dy = 0, 0
    for dyc in range(int(min_height), int(dyc_max) + 1):
        # We can go as far as (a) the first exclusion box OR
        # (b) the first inclusion box that crosses at the
        # starting (yc) or current (yc + dyc) y-value or (c)
        # the maximum width for the boxes or (d) the width
        # of the image.
        dxc_max = np.concatenate(
            [
                xye[(xye[:, 1] < (y + dyc)) & (xye[:, 0] != x), 0] - x,
                xyi[
                    ((xyi[:, 1] < (y + dyc)) & (xyi[:, 3] > (y + dyc)))
                    | (xyi[:, 1] < y),
                    0,
                ]
                - x,
                [max_width],
            ]
        ).min()
        # Ranges of x-values that would result in splitting an
        # inclusion box.
        inclusion_ranges = xyi[(xyi[:, 1] < (y + dyc))][:, [0, 2]]
        for dxc in range(int(dxc_max), int(max(min_width - 1, 0)), -1):
            if (
                len(inclusion_ranges) == 0
                or (
                    # The inclusion range starts after this point.
                    (inclusion_ranges[:, 0] >= (x + dxc))
                    # The inclusion range ends before this point.
                    | (inclusion_ranges[:, 1] <= (x + dxc))
                ).all()
            ):
                if (dxc * dyc) >= (dx * dy):
                    dx, dy = dxc, dyc
                break
    return dx, dy, (dx > 0 and dy > 0)


def find_acceptable_crops(
    include: np.ndarray,
    width: int,
    height: int,
    exclude: np.ndarray = None,
    max_width: int = None,
    max_height: int = None,
    cache=None,
) -> np.ndarray:
    """Given a list of consensus and non-consensus regions, crop the image into segments
    that avoid the non-consensus regions while not splitting the consensus regions."""
    if max_width is None:
        max_width = width
    if max_height is None:
        max_height = height
    if exclude is None:
        exclude = np.empty((0, 4))
    key = None
    if cache is not None:
        key = (
            hash(include.tobytes()),
            width,
            height,
            hash(exclude.tobytes()),
            max_width,
            max_height,
        )
        crops = cache.get(key)
        if crops is not None:
            return crops
    crops = []
    yfrontier = np.zeros(width, dtype="int32")
    while True:
        xc1 = typing.cast(int, yfrontier.argmin())
        yc1 = yfrontier[xc1]
        dx1, dy1, success1 = search(
            x=xc1,
            y=yc1,
            exclude=exclude,
            include=include,
            max_height=min(max_height, height - yc1),
            max_width=min(max_width, width - xc1),
        )
        xc2, yc2 = xc1 + dx1, yc1 + dy1
        if (yfrontier[xc1:xc2] == yc2).all():
            # We've made no progress. Stop.
            break
        yfrontier[xc1:xc2] = yfrontier[xc1:xc2].clip(min=yc2)
        if not success1:
            continue
        dx2, dy2, success2 = search(
            x=width - (xc1 + dx1),
            y=height - (yc1 + dy1),
            exclude=(width, height, width, height) - exclude[:, [2, 3, 0, 1]],
            include=(width, height, width, height) - include[:, [2, 3, 0, 1]],
            max_height=min(max_height, yc1 + dy1),
            max_width=min(max_width, xc1 + dx1),
            min_height=dy1,
            min_width=dx1,
        )
        if success2:
            xc1, yc1 = xc2 - dx2, yc2 - dy2
            yfrontier[xc1:xc2] = yfrontier[xc1:xc2].clip(min=yc2)
        crops.append([xc1, yc1, xc2, yc2])
    crops = np.array(crops).round().astype("int32")
    if cache is not None:
        cache[key] = crops
    return crops


def visualize_crops(
    include: np.ndarray,
    crops: np.ndarray,
    width: int,
    height: int,
    exclude: np.ndarray = None,
    canvas: np.ndarray = None,
) -> np.ndarray:
    """Create a visual of the consensus, non-consensus, and a set of crops."""
    if exclude is None:
        exclude = np.empty((0, 4))
    visual = (
        canvas
        if canvas is not None
        else (np.zeros((height, width, 4)) + (0, 0, 0, 255)).astype("uint8")
    )
    if canvas is None:
        # Draw filled regions if it's a blank image.
        for xc1, yc1, xc2, yc2 in crops:
            cv2.rectangle(
                visual,
                pt1=(xc1, yc1),
                pt2=(xc2, yc2),
                thickness=-1,
                color=(0, 0, 255, 120),
            )
    for xc1, yc1, xc2, yc2 in crops:
        cv2.rectangle(
            visual,
            pt1=(xc1, yc1),
            pt2=(xc2, yc2),
            thickness=2,
            color=(255, 255, 0, 255),
        )
    for x1, y1, x2, y2 in include:
        visual[y1:y2, x1:x2] = (0, 255, 0, 255)
    for x1, y1, x2, y2 in exclude:
        visual[y1:y2, x1:x2] = (255, 0, 0, 255)
    return visual



import warnings

warnings.filterwarnings(
    action="ignore",
    message=r".*The output shape of `ResNet50\(include_top=False\)` has been changed since Keras 2\.2\.0\..*",
    category=UserWarning,
)
__version__ = "0.5.30"
